package ch.psi.daq.queryrest.config;

import java.util.Arrays;
import java.util.EnumMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.function.Function;

import javax.annotation.Resource;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Import;
import org.springframework.context.annotation.PropertySource;
import org.springframework.core.env.Environment;
import org.springframework.http.converter.HttpMessageConverter;
import org.springframework.http.converter.json.MappingJackson2HttpMessageConverter;
import org.springframework.util.StringUtils;
import org.springframework.validation.Validator;
import org.springframework.web.servlet.config.annotation.WebMvcConfigurerAdapter;

import com.fasterxml.jackson.annotation.JsonInclude.Include;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.Version;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.module.SimpleModule;

import ch.psi.daq.cassandra.util.test.CassandraDataGen;
import ch.psi.daq.common.json.deserialize.AttributeBasedDeserializer;
import ch.psi.daq.common.statistic.StorelessStatistics;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.query.analyzer.QueryAnalyzer;
import ch.psi.daq.query.analyzer.QueryAnalyzerImpl;
import ch.psi.daq.query.config.QueryConfig;
import ch.psi.daq.query.model.Aggregation;
import ch.psi.daq.query.model.Query;
import ch.psi.daq.query.model.QueryField;
import ch.psi.daq.query.model.impl.AbstractQuery;
import ch.psi.daq.queryrest.controller.validator.QueryValidator;
import ch.psi.daq.queryrest.model.PropertyFilterMixin;
import ch.psi.daq.queryrest.response.JsonByteArraySerializer;
import ch.psi.daq.queryrest.response.JsonStreamSerializer;
import ch.psi.daq.queryrest.response.ResponseStreamWriter;

@Configuration
@PropertySource(value = {"classpath:queryrest.properties"})
@PropertySource(value = {"file:${user.home}/.config/daq/queryrest.properties"}, ignoreResourceNotFound = true)
public class QueryRestConfig extends WebMvcConfigurerAdapter {

   private static final String QUERYREST_DEFAULT_RESPONSE_AGGREGATIONS = "queryrest.default.response.aggregations";

   private static final String QUERYREST_DEFAULT_RESPONSE_FIELDS = "queryrest.default.response.fields";

   // a nested configuration
   // this guarantees that the ordering of the properties file is as expected
   // see:
   // https://jira.spring.io/browse/SPR-10409?focusedCommentId=101393&page=com.atlassian.jira.plugin.system.issuetabpanels:comment-tabpanel#comment-101393
   @Configuration
   @Import({QueryConfig.class})
   static class InnerConfiguration {
   }

   private static final Logger LOGGER = LoggerFactory.getLogger(QueryRestConfig.class);

   public static final String BEAN_NAME_DEFAULT_RESPONSE_FIELDS = "defaultResponseFields";
   public static final String BEAN_NAME_DEFAULT_RESPONSE_AGGREGATIONS = "defaultResponseAggregations";

   @Resource
   private Environment env;

   /**
    * {@inheritDoc}
    */
   @Override
   public void configureMessageConverters(List<HttpMessageConverter<?>> converters) {
      final MappingJackson2HttpMessageConverter converter = new MappingJackson2HttpMessageConverter();
      /**
       * This is necessary so that the message conversion uses the configured object mapper.
       * Otherwise, a separate object mapper is instantiated for Springs message conversion.
       */
      converter.setObjectMapper(objectMapper());
      converters.add(converter);
      super.configureMessageConverters(converters);
   }

   @Bean
   public ObjectMapper objectMapper() {
      ObjectMapper mapper = new ObjectMapper();

      String packageName = AbstractQuery.class.getPackage().getName();
      Class<AbstractQuery> abstractQueryClass = AbstractQuery.class;
      AttributeBasedDeserializer<AbstractQuery> abstractQueryDeserializer =
            new AttributeBasedDeserializer<AbstractQuery>(abstractQueryClass).register(packageName);
      SimpleModule module = new SimpleModule("PolymorphicAbstractQuery", Version.unknownVersion());
      module.addDeserializer(abstractQueryClass, abstractQueryDeserializer);
      mapper.registerModule(module);

      // only include non-null values
      mapper.setSerializationInclusion(Include.NON_NULL);
      // Mixin which is used dynamically to filter out which properties get serialised and which
      // won't. This way, the user can specify which columns are to be received.
      mapper.addMixIn(DataEvent.class, PropertyFilterMixin.class);
      mapper.addMixIn(StorelessStatistics.class, PropertyFilterMixin.class);
      mapper.addMixIn(EnumMap.class, PropertyFilterMixin.class);

      // defines how to writer inner Streams (i.e. Stream<Entry<String, Stream<?>>> toSerialize)
      module = new SimpleModule("Streams API", Version.unknownVersion());
      module.addSerializer(new JsonStreamSerializer());
      mapper.registerModule(module);

      // by default, byte[] are written as String
      module = new SimpleModule("byte[]", Version.unknownVersion());
      module.addSerializer(new JsonByteArraySerializer());
      mapper.registerModule(module);

      return mapper;
   }

   @Bean
   public JsonFactory jsonFactory() {
      return new JsonFactory();
   }

   @Bean
   public Function<Query, QueryAnalyzer> queryAnalizerFactory() {
      return (query) -> new QueryAnalyzerImpl(query);
   }

   @Bean
   public ResponseStreamWriter responseStreamWriter() {
      return new ResponseStreamWriter();
   }

   @Bean(name = BEAN_NAME_DEFAULT_RESPONSE_FIELDS)
   public Set<QueryField> defaultResponseFields() {
      String[] responseFields =
            StringUtils.commaDelimitedListToStringArray(env.getProperty(QUERYREST_DEFAULT_RESPONSE_FIELDS));
      LOGGER.debug("Load '{}={}'", BEAN_NAME_DEFAULT_RESPONSE_FIELDS, Arrays.toString(responseFields));

      // preserve order
      LinkedHashSet<QueryField> defaultResponseFields = new LinkedHashSet<>(responseFields.length);
      for (String field : responseFields) {
         try {
            defaultResponseFields.add(QueryField.valueOf(field));
         } catch (Exception e) {
            LOGGER.error("Field '{}' in '{}' is invalid.", field, QUERYREST_DEFAULT_RESPONSE_FIELDS, e);
         }
      }

      return defaultResponseFields;
   }

   @Bean(name = BEAN_NAME_DEFAULT_RESPONSE_AGGREGATIONS)
   public Set<Aggregation> defaultResponseAggregations() {
      String[] responseAggregations =
            StringUtils.commaDelimitedListToStringArray(env.getProperty(QUERYREST_DEFAULT_RESPONSE_AGGREGATIONS));
      LOGGER.debug("Load '{}={}'", BEAN_NAME_DEFAULT_RESPONSE_AGGREGATIONS, Arrays.toString(responseAggregations));

      // preserve order
      LinkedHashSet<Aggregation> defaultResponseAggregations = new LinkedHashSet<>(responseAggregations.length);
      for (String aggregation : responseAggregations) {
         try {
            defaultResponseAggregations.add(Aggregation.valueOf(aggregation));
         } catch (Exception e) {
            LOGGER.error("Aggregation '{}' in '{}' is invalid.", aggregation, QUERYREST_DEFAULT_RESPONSE_AGGREGATIONS,
                  e);
         }
      }

      return defaultResponseAggregations;
   }

   @Bean
   public Validator queryValidator() {
      return new QueryValidator();
   }

   // ==========================================================================================
   // TODO: This is simply for initial / rudimentary testing - remove once further evolved
   @Bean
   public CassandraDataGen cassandraDataGen() {
      return new CassandraDataGen();
   }
}
