package ch.psi.daq.queryrest.controller;

import java.io.IOException;
import java.util.Collection;
import java.util.Map.Entry;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Stream;

import javax.annotation.Resource;
import javax.servlet.http.HttpServletResponse;
import javax.validation.Valid;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.MediaType;
import org.springframework.validation.Validator;
import org.springframework.web.bind.WebDataBinder;
import org.springframework.web.bind.annotation.InitBinder;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.RestController;

import ch.psi.daq.cassandra.request.validate.RequestProviderValidator;
import ch.psi.daq.cassandra.util.test.CassandraDataGen;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.query.analyzer.QueryAnalyzer;
import ch.psi.daq.query.model.Aggregation;
import ch.psi.daq.query.model.DBMode;
import ch.psi.daq.query.model.Query;
import ch.psi.daq.query.model.QueryField;
import ch.psi.daq.query.model.impl.AbstractQuery;
import ch.psi.daq.query.processor.QueryProcessor;
import ch.psi.daq.queryrest.config.QueryRestConfig;
import ch.psi.daq.queryrest.model.ChannelsRequest;
import ch.psi.daq.queryrest.response.ResponseStreamWriter;

@RestController
public class QueryRestController {

   private static final Logger LOGGER = LoggerFactory.getLogger(QueryRestController.class);

   public static final String CHANNELS = "/channels";
   public static final String QUERY = "/query";

   @Resource
   private Validator queryValidator;
   private Validator requestProviderValidator = new RequestProviderValidator();

   @Resource
   private ResponseStreamWriter responseStreamWriter;

   @Resource
   private QueryProcessor cassandraQueryProcessor;

   @Resource
   private QueryProcessor archiverApplianceQueryProcessor;

   @Resource
   private Function<Query, QueryAnalyzer> queryAnalizerFactory;

   @Resource(name = QueryRestConfig.BEAN_NAME_DEFAULT_RESPONSE_FIELDS)
   private Set<QueryField> defaultResponseFields;

   @Resource(name = QueryRestConfig.BEAN_NAME_DEFAULT_RESPONSE_AGGREGATIONS)
   private Set<Aggregation> defaultResponseAggregations;


   @InitBinder
   protected void initBinder(WebDataBinder binder) {
      if (binder.getTarget() != null) {
         if (requestProviderValidator.supports(binder.getTarget().getClass())) {
            binder.addValidators(requestProviderValidator);
         }
         if (queryValidator.supports(binder.getTarget().getClass())) {
            binder.addValidators(queryValidator);
         }
      }
   }

   @RequestMapping(
         value = CHANNELS,
         method = {RequestMethod.GET, RequestMethod.POST},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody Collection<String> getChannels(@RequestBody(required = false) ChannelsRequest request)
         throws Throwable {
      // in case not specified use default (e.g. GET)
      if (request == null) {
         request = new ChannelsRequest();
      }

      try {
         return getQueryProcessor(request.getDbMode()).getChannels(request.getRegex());
      } catch (Throwable t) {
         LOGGER.error("Failed to query channel names.", t);
         throw t;
      }
   }

   @RequestMapping(
         value = QUERY,
         method = RequestMethod.POST,
         consumes = {MediaType.APPLICATION_JSON_VALUE})
   public void executeQuery(@RequestBody @Valid AbstractQuery query, HttpServletResponse res) throws IOException {
      try {
         LOGGER.debug("Executing query '{}'", query.getClass().getSimpleName());

         QueryAnalyzer queryAnalizer = queryAnalizerFactory.apply(query);

         // all the magic happens here
         Stream<Entry<String, Stream<? extends DataEvent>>> channelToDataEvents =
               getQueryProcessor(query.getDbMode()).process(queryAnalizer);

         // do post-process
         Stream<Entry<String, ?>> channelToData = queryAnalizer.postProcess(channelToDataEvents);

         // write the response back to the client using java 8 streams
         responseStreamWriter.respond(channelToData, query, res);
      } catch (Throwable t) {
         LOGGER.error("Failed execute query '{}'.", query, t);
         throw t;
      }
   }

   private QueryProcessor getQueryProcessor(DBMode dbMode) {
      if (DBMode.databuffer.equals(dbMode)) {
         return cassandraQueryProcessor;
      } else if (DBMode.archiverappliance.equals(dbMode)) {
         return archiverApplianceQueryProcessor;
      } else {
         LOGGER.error("Unknown DBMode '{}'!", dbMode);
         throw new IllegalArgumentException(String.format("Unknown DBMode '%s'", dbMode));
      }
   }

   // ==========================================================================================
   // TODO: This is simply for initial / rudimentary testing - remove once further evolved

   @Resource
   private CassandraDataGen cassandraDataGen;

   @RequestMapping(value = "/write")
   public void writeDummyEntry() {
      long nrOfElements = 4;
      cassandraDataGen.writeData(3, 0, 4,
            i -> i,
            i -> 0,
            i -> i,
            i -> i,
            i -> 0,
            i -> new long[] {i, i, i, i},
            "TRFCA-channel1");

      cassandraDataGen.writeData(3, 0, 4,
            i -> i,
            i -> 0,
            i -> i,
            i -> i,
            i -> 0,
            i -> new long[] {nrOfElements - i, nrOfElements - i, nrOfElements - i, nrOfElements - i},
            "TRFCA-channel2");

      cassandraDataGen.writeData(3, 0, 4,
            i -> i,
            i -> 0,
            i -> i,
            i -> i,
            i -> 0,
            i -> i,
            "TRFCB-channel3");

      cassandraDataGen.writeData(3, 0, 4,
            i -> i,
            i -> 0,
            i -> i,
            i -> i,
            i -> 0,
            i -> nrOfElements - i,
            "TRFCB-channel4");
   }
}
