package ch.psi.daq.queryrest.response;

import java.io.IOException;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.Map.Entry;
import java.util.stream.Stream;

import javax.annotation.Resource;
import javax.servlet.ServletResponse;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.MediaType;

import com.fasterxml.jackson.core.JsonEncoding;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.ObjectWriter;
import com.fasterxml.jackson.databind.ser.impl.SimpleBeanPropertyFilter;
import com.fasterxml.jackson.databind.ser.impl.SimpleFilterProvider;

import ch.psi.daq.query.model.Aggregation;
import ch.psi.daq.query.model.QueryField;
import ch.psi.daq.query.model.impl.AbstractQuery;

/**
 * Takes a Java 8 stream and writes it to the output stream provided by the {@link ServletResponse}
 * of the current request.
 */
public class ResponseStreamWriter {

   private static final Logger logger = LoggerFactory.getLogger(ResponseStreamWriter.class);

   @Resource
   private JsonFactory jsonFactory;

   @Resource
   private ObjectMapper mapper;

   /**
    * Responding with the the contents of the stream by writing into the output stream of the
    * {@link ServletResponse}.
    * 
    * @param stream Mapping from channel name to data
    * @param query concrete instance of {@link AbstractQuery}
    * @param response {@link ServletResponse} instance given by the current HTTP request
    * @throws IOException thrown if writing to the output stream fails
    */
   public void respond(Stream<Entry<String, ?>> stream, AbstractQuery query,
         ServletResponse response) throws IOException {

      Set<QueryField> queryFields = query.getFields();
      List<Aggregation> aggregations = query.getAggregations();

      Set<String> includedFields =
            new LinkedHashSet<String>(queryFields.size() + (aggregations != null ? aggregations.size() : 0));

      for (QueryField field : queryFields) {
         includedFields.add(field.name());
      }
      if (aggregations != null) {
         for (Aggregation aggregation : query.getAggregations()) {
            includedFields.add(aggregation.name());
         }
      }
      // do not write channel since it is already provided as key in mapping
      includedFields.remove(QueryField.channel.name());

      ObjectWriter writer = configureWriter(includedFields);
      respondInternal(stream, response, writer);
   }

   /**
    * Configures the writer dynamically by including the fields which should be included in the
    * response.
    * 
    * @param includedFields set of strings which correspond to the getter method names of the
    *        classes registered as a mixed-in
    * @return the configured writer that includes the specified fields
    */
   private ObjectWriter configureWriter(Set<String> includedFields) {
      SimpleFilterProvider propertyFilter = new SimpleFilterProvider();
      propertyFilter.addFilter("namedPropertyFilter", SimpleBeanPropertyFilter.filterOutAllExcept(includedFields));
      // only write the properties not excluded in the filter
      ObjectWriter writer = mapper.writer(propertyFilter);
      return writer;
   }
   
   /**
    * Writes the outer Java stream into the output stream.
    * 
    * @param stream Mapping from channel name to data
    * @param response {@link ServletResponse} instance given by the current HTTP request
    * @param writer configured writer that includes the fields the end user wants to see
    * @throws IOException thrown if writing to the output stream fails
    */
   private void respondInternal(Stream<Entry<String, ?>> stream, ServletResponse response,
         ObjectWriter writer)
         throws IOException {

      response.setCharacterEncoding(JsonEncoding.UTF8.getJavaName());
      response.setContentType(MediaType.APPLICATION_JSON_VALUE);
      JsonGenerator generator = jsonFactory.createGenerator(response.getOutputStream(), JsonEncoding.UTF8);
      generator.writeStartArray();
      stream
            /* ensure elements are sequentially written */
            .sequential()
            .forEach(
                  entry -> {
                     try {
                        generator.writeStartObject();
                        generator.writeStringField(QueryField.channel.name(), entry.getKey());

                        generator.writeFieldName("data");
                        writer.writeValue(generator, entry.getValue());
                        
                        generator.writeEndObject();
                     } catch (Exception e) {
                        logger.error("Could not write channel name of channel '{}'", entry.getKey(), e);
                     }
                  });
      generator.writeEndArray();
      generator.flush();
      generator.close();
   }

}
